/* 
 * File:   main.cpp
 * Author: Will Matheson
 * Course: PROG 2400 (Data Structures), Section 700
 * Due:    March 25, 2013
 * 
 * PROG 2400 Assignment #3: Searching
 * Part 1: Sequential and Binary Searching
 * 
 * Searches for letters A-Z in a static sorted array.
 * Captures and displays search times for the
 * sequential searches and the binary searches.
 *
 * Created on February 28, 2013, 1:57 PM
 */

#include <cstdlib>
#include <iostream>
#include "Timer.h"

using namespace std;

/*
 * Driver and functions for program.
 */

//fast integer log2, for computing binary search worst-case
//(what you need to do to 2 (2^?) to get a given value)
//http://stackoverflow.com/a/994647
unsigned int lg (unsigned int val) {
    unsigned int retVal = -1;//if given 0, but mathematical answer is infinity
    while (val != 0) {
        val >>= 1;//shift value to the right by one bit
        retVal++;//however many times that had to be done is what we want
        //since it represents the power of 2 to get the value
    }
    return retVal;
}

unsigned long binarySearch(char searchArray[],
        int arraySize, char searchItem)
{
   
    int searchPoint = (arraySize / 2) - 1; //zero-indexed array
    int searchIncrement = arraySize / 4;
    bool found = false;
    int maxSearchIterations = lg(arraySize) + 1; //worstcase of lg2(arraySize)+1
    int iterations = 0;
    
    //start the timer
    Timer* myTimer = new Timer();
    myTimer->start();
    
    while(!found && iterations > maxSearchIterations)
    {
        if (searchArray[searchPoint] < searchItem)
        {
            //if we were too low, make the search point jump ahead
            searchPoint += searchIncrement;
        }
        else if (searchArray[searchPoint] > searchItem)
        {
            //if we were too high, make the search point jump back
            searchPoint -= searchIncrement;                    
        }
        else
        {
            //we found it!
            found = true;       
        }
        if (searchIncrement > 1)
        {
            //make the search increment smaller (if possible)
            searchIncrement = searchIncrement / 2;
        }
        iterations++;
    }
    
    //return time elapsed
    unsigned long elapsedTime = myTimer->elapsedTimeNanoseconds();
    delete myTimer;
    return elapsedTime;
}

unsigned long sequentialSearch(char searchArray[],
        int arraySize, char searchItem)
{
    //start the timer
    Timer* myTimer = new Timer();
    myTimer->start();
    unsigned long elapsedTime;
    
    //find the character
    for (int i=0; i< arraySize; i++)
    {
        if (searchArray[i] == searchItem)
        {
            //return elapsed time (when found)
            elapsedTime = myTimer->elapsedTimeNanoseconds();
            delete myTimer;
            return elapsedTime;
        }
    }
    
    //return elapsed time (didn't find it)
    elapsedTime = myTimer->elapsedTimeNanoseconds();
    delete myTimer;
    return elapsedTime;
}

int main(int argc, char** argv) {
    //array to contain letters A-Z
    const int arraySize = 26;
    static char alphabet[arraySize];
    //store search times
    unsigned long sequientialSearchTimes[arraySize];
    unsigned long binarySearchTimes[arraySize];
    //character counter, for going from A-Z
    char current = 'A';
    
    //search and fill in results
    cout << "Search Times (ns):" << endl;
    cout << "Letter | Sequential | Binary" << endl;
    
    //it is possible to do all of the following in one loop:
    /*for (int i=0; i<arraySize; i++)
    {
        //build up the alphabet as we go
        alphabet[i] = current;
        //do the sequential, binary searches, outputting the results
        cout << current << " "
                << sequentialSearch(alphabet, arraySize, current) << " "
                << binarySearch(alphabet, arraySize, current) << endl;
        current++;
    }*/
    
    //but this is easier to tinker with:
    
    //build up the alphabet
    for (int i=0; i<arraySize; i++)
    {
        alphabet[i] = current;
        current++;
    }
    
    //do sequential searches
    for (int i=0; i<arraySize; i++)
    {
        sequientialSearchTimes[i] =
                sequentialSearch(alphabet, arraySize, alphabet[i]);
    }

    //do binary searches
    for (int i=0; i<arraySize; i++)
    {
        binarySearchTimes[i] =
                binarySearch(alphabet, arraySize, alphabet[i]);
    }
    
    //output results
    for (int i=0; i<arraySize;i++)
    {
    cout << "   " << alphabet[i] << "        " << sequientialSearchTimes[i]
            << "        "  << binarySearchTimes[i] << endl;
    }

    return 0;
}

