/* 
 * File:   SpellChecker.cpp
 * Author: Will Matheson
 * Course: PROG 2400 (Data Structures), Section 700
 * Due:    March 25, 2013
 * 
 * Implementation for SpellChecker (see header file for description)
 * 
 * Created on March 13, 2013, 3:06 PM
 */

#include "SpellChecker.h"

//**BIG THREE**
//constructor (takes document and dictionary filenames)
SpellChecker::SpellChecker(std::string documentFileName,
        std::string dictionaryFileName) {
    _documentFileName = documentFileName;
    _dictionaryFileName = dictionaryFileName;
    _dictionary = new Tree();
    //balancing demonstration:
    _dictionary->setDemoMode(true);
    _documentWords = new Stack();
    _misspelledWords = new LinkedList();
}

//copy constructor (not defined)
SpellChecker::SpellChecker(const SpellChecker& orig) {
}

//virtual destructor (not defined)
SpellChecker::~SpellChecker() {
}

//**PUBLIC UTILITIES**
//check spelling in document
void SpellChecker::checkSpelling() {
    //read all the words from the dictionary file and put them in the tree
    //(also balance the tree)
    if (!loadDictionary())
    {
        return;//dictionary failed to load
    }
    
    //display dictionary for review
    displayDictionary();
    
    //read and store what are really words from the document file
    if (!loadDocumentWords())
    {
        return;//document failed to load    
    }
    
    //display the potentially misspelled words
    if (anyMisspelledWords())
    {
        std::cout << "Your potentially misspelled words are: " << std::endl;
        std::cout << *_misspelledWords;    
    }
    else
    {
        std::cout << "No potential misspellings were found." << std::endl;    
    }   
}

//"cleans" a word of special characters
//http://stackoverflow.com/q/1315041
//http://www.cplusplus.com/forum/beginner/12207/
std::string SpellChecker::wordClean(std::string word){
    std::string::iterator it;//a way to do things at a point in a string
    int counter = 0;//just here for syntax purposes
    //going from the beginning of the word to the end of the word
    //note that value of word.end() can change if we're knocking off characters!
    for (it = word.begin(); it < word.end(); counter++)
    {
        //if the character is a special character...
        if ((*it) == '.'
                ||(*it) == '"'
                ||(*it) == ','
                ||(*it) == '&'
                ||(*it) == '('
                ||(*it) == ')'
                ||(*it) == '#')
        {
            //zap it!
            it = word.erase (it);
        }
        else //go on to the next character
        {
            it++;
        }
    }
    return word;
}

//returns true if a word is only digits
//http://stackoverflow.com/a/4654718
bool SpellChecker::isNumericWord(std::string word){
    std::string::iterator it = word.begin();//way to access a point in a string
    //keep iterator going as long as we haven't hit the end AND
    //(the character at the point of the iterator is a digit OR
    //perhaps the sign of an integer)
    while (it != word.end() && (std::isdigit(*it)
            || (*it) == '-' || (*it) == '+'))
    {
        ++it;
    }
    //both have to be true to evaluate to true
    //if the iterator didn't make it through, it wasn't a "numeric" word
    return (!word.empty() && it == word.end());
}

//convert a string to lowercase
//http://www.codepedia.com/1/CppToLowerCase
std::string SpellChecker::convertToLower(std::string word){
    //the length isn't going to change, so we don't need an iterator
    const int length = word.length();
    for(int i=0; i!=length; ++i)
    {
        word[i] = std::tolower(word[i]);
    }
    return word;
}

//**INTERNAL FUNCTIONS**
//load dictionary file into BST, return true if successful
bool SpellChecker::loadDictionary(){
    std::string word;//hold a dictionary word
    
    //does the file exist?
    //declare and open stream (takes old-school string)
    std::ifstream inStream(_dictionaryFileName.c_str());
    if (!inStream.fail())
    {
        while (!inStream.eof())
        {
            //grab a word (one word per line)
            getline(inStream,word);
            //and stick it in the BST
            if (word != "")
            {
                _dictionary->insert(word);
            }
        }
    }
    else
    {
        std::cout << "Failed to load dictionary." << std::endl;
        inStream.close();
        return false;
    }
    //close the stream
    inStream.close();
    //balance BST
    _dictionary->rebalance();
    return true;
}

//display dictionary BST
void SpellChecker::displayDictionary(){
    std::cout << "The dictionary file you loaded contains the following words:"
            << std::endl;
    std::cout << *_dictionary; //alphabetical order
    std::cout << "The words are organized as follows:"
            << std::endl;
    _dictionary->setLevelOrder(true);
    std::cout << *_dictionary; //level-order
}

//load document words into queue, return true if successful
bool SpellChecker::loadDocumentWords(){
    std::string word;//hold a document word
    
    //does the file exist?
    //declare and open stream (takes old-school string)
    std::ifstream inStream(_documentFileName.c_str());
    if (!inStream.fail())
    {
        //grabs one (space separated) word at a time
        while (inStream >> word)
        {
            //clean out the word
            word = wordClean(word);
            //if anything left, and it's not a representation of a number,
            //put it on the list of words to check
            if (word != "" && !isNumericWord(word))
            {
                _documentWords->push(word);
            }
        }
    }
    else
    {
        std::cout << "Failed to load document." << std::endl;
        inStream.close();
        return false;
    }
    //close the stream
    inStream.close();
    return true;
}

//searches dictionary, returns true if any words misspelled
bool SpellChecker::anyMisspelledWords(){
    bool anyMSW = false;//"any misspelled words"
    std::string word;//store a word from the stack
    std::string lcaseWord;//lowercase version to test against dictionary
    
    //as long as we have potential words to check
    while (!_documentWords->isEmpty())
    {
        //grab the next word off the stack
        word = _documentWords->slip();
        //convert it to lowercase
        lcaseWord = convertToLower(word);
        //is it in the dictionary? (which is all lowercase)
        if (!_dictionary->found(lcaseWord))
        {
            //put the potentially misspelled word on the list
            //note that we're checking lowercase versions of the words
            //but in the end showing the word in its original casing
            _misspelledWords->append(word);
            anyMSW = true;
        }
    }
    
    return anyMSW;
}
